#!/usr/bin/env tclsh

############################################################

#
# rbc.tcl -- Tcl script to conduct and optimize blastclust
#     clustering. Tested with tclsh version 8.4 and
#     blastclust version 2.2.10 and 2.2.17
#
# (C) 2007/2008 by
#     Markus Goeker (support [AT] goeker [DOT] org)
#
# This program is distributed under the terms of the Gnu
# Public License V2. For further information, see
# http://www.gnu.org/licenses/gpl.html
#
# If you use this script in a publication, please cite the
# web page at http://www.goeker.org/mg/clustering/
#

############################################################

# Run blastclust and return a list of lists which represent the
# clusters. Input is an array that stores the blastclust options.
proc runBC {argArr {bcprog blastclust}} {
    upvar $argArr argArrName
    set clusters {}
    if {[catch {
        set blastout [open "|$bcprog [array get argArrName]" r]
        while {[gets $blastout line] >= 0} {
            if {[string first " Start clustering of " $line] >= 0} {
                continue
            }
            lappend clusters [lrange [split $line " "] 0 end-1]
        }
        close $blastout
        file delete error.log
    } problem]} {
        error "$problem\nBlastclust did not result."
    }
    return $clusters
}

##################################################

# Include group affiliation numbers from current partition in an
# array that stores the results from all partitions, using the
# reference number "runno".
proc appendclusters {outarr runno partition} {
    upvar $outarr outarrName
    set clustno 0
    foreach cluster $partition {
        foreach taxon $cluster {
            set outarrName($taxon,$runno) $clustno
        }
        incr clustno
    }
    return
}

##################################################

# Flatten and sort a nested list
proc gettaxa partition {
    proc flatten inlist {
        set taxa {}
        foreach i $inlist {
            if {[llength $i] > 1} {
                append taxa [flatten $i] " "
            } else {
                append taxa [lindex $i 0] " "
            }
        }
        return $taxa
    }
    return [lsort [flatten $partition]]
}

##################################################

# Convert a tcl list to a Newick tree. The terminal
# semicolon is not appended!
proc listToNewick inlist {
    set tree ""
    set count 0
    foreach i $inlist {
        if {$count} {
            append tree ","
        }
        if {[llength $i] > 1} {
            append tree [listToNewick $i] ":1.0"
        } else {
            append tree [lindex $i 0] ":1.0"
        }
        incr count
    }
    return "($tree)"
}

##################################################

# Convert a tcl list to a CSV table.
proc listToCSV inlist {
    set grpno 0
    set result ""
    foreach group $inlist {
        incr grpno
        foreach taxon $group {
            append result "$taxon\t$grpno\n"
        }
    }
    return $result
}

##################################################

# Read settings from file.
proc readSetFile filename {
    if {[catch {
        set infile [open $filename r]
        while {[gets $infile line] >= 0} {
            if {[string is space $line] || [string match "#*" $line]} {
                continue
            }
            set part [split $line ":"]
            set varName [string trim [lindex $part 0]]
            if {! [regexp "(min|max|step)(Score|Word|Cover)|bothCover" \
                    $varName]} {
                error "Unknown parameter: \"$varName\""
            }
            if {[info exists checkArr($varName)]} {
                error "2nd occurrence of parameter \"$varName\"."
            }
            set checkArr($varName) 1
            upvar $varName tmp
            set tmp [string trim [lindex $part 1]]
        }
        close  $infile
    } problem]} {
        error $problem
    }
    return
}

##################################################

# Read settings from command line.
proc bcset {arg bcArr} {
    upvar $bcArr bcArrName
    foreach value [lrange [split $arg ":"] 1 end] name {-S -W -L -b} {
        if {[string is space $value]} {
            continue
        }
        set bcArrName($name) [string trim $value]
    }
    return
}

##################################################

# Print error message.
proc usage {} {
    global argv0
    set program [file tail $argv0]
    puts stderr "Script to perform and optimize blastclust clustering."
    puts stderr ""
    puts stderr "Usage:"
    puts stderr ""
    puts stderr "Clustering under default values (except for -b F):"
    puts stderr "  $program fastafile"
    puts stderr ""
    puts stderr "Customized clustering:"
    puts stderr "  $program fastafile :score:word:length:both"
    puts stderr "  (e.g. $program seq.fas :95:48:0.8:T)"
    puts stderr ""
    puts stderr "Optimization of clustering:"
    puts stderr "  $program fastafile settingsfile"
    puts stderr "  (\"settingsfile\" must not start with \":\""
    puts stderr ""
    puts stderr "Author: Markus Goeker. Contact: <support@goeker.org>"
    puts stderr ""
    return
}

##################################################

# Parse command line

switch -exact -- [llength $argv] {
    1 { ;# Newick mode, default settings
        set runmode newick
        set bcarg(-b) F
    }
    2 { ;# Newick mode, settings read from command line
        if {[string index [lindex $argv 1] 0] == ":"} {
            set runmode newick
            set bcarg(-b) F
            if {[catch {bcset [lindex $argv 1] bcarg} problem]} {
                puts stderr $problem
                exit 1
            }
        } else { ;# Optimization mode, settings read from file
            set runmode optimize
            set setfile [lindex $argv 1]
        }
    }
    default { ;# Help message
        usage
        exit 1
    }
}

##################################################

# Set the constants:

# Fasta file
set bcarg(-i) [lindex $argv 0]
if {! [file exists $bcarg(-i)]} {
    puts stderr "\"$bcarg(-i)\": no such file or directory"
    exit 1
}

# Nucleotide sequences
set bcarg(-p) F

# Temporary storage of sequence neighbours
set neighboursfile neighbours_[lindex [split [expr {rand()}] .] 1]

##################################################

# Newick mode -- compute clustering and quit:

if {$runmode == "newick"} {
    if {[catch {
        set bcList [runBC bcarg]
        puts "[listToNewick $bcList];"
        puts ""
        puts "[listToCSV $bcList]"
        exit
    } problem]} {
        puts stderr $problem
        exit 1
    }
}

##################################################

# Optimization mode:

if {[catch {

    # Read settings
    readSetFile $setfile

    # Main loop
    set runNo 0

    # Word sizes
    for {set k $minWord} {$k <= $maxWord} {set k [expr {$k + $stepWord}]} {

        if {$k == $minWord} {
            set firstRun 1
        } else {
            set firstRun 0
        }
        set bcarg(-W) $k

        for {set j $minCover} {$j <= $maxCover} {set j [expr {$j + \
                $stepCover}]} {

            set bcarg(-L) $j

            # Coverage required in both neighbours?
            foreach i $bothCover {

                set bcarg(-b) $i

                # Score coverage thresholds
                for {set l $minScore} {$l <= $maxScore} {set l [expr {$l + \
                        $stepScore}]} {

                    # In the 1st run, we write the information on neighbours
                    # to "$neighboursfile"; in all subsequent runs, we read
                    # from that file. Because re-running megablast is avoided,
                    # this causes a considerable speedup.
                    if {$firstRun} {
                        set bcarg(-s) $neighboursfile
                        set firstRun 0
                    } else {
                        unset -nocomplain bcarg(-s)
                        set bcarg(-r) $neighboursfile
                    }

                    set bcarg(-S) $l
                    set bcList [runBC bcarg]
                    lappend scores [expr {$l < 3.0 ? $l : $l / 100.0}]
                    appendclusters results $runNo $bcList 
                    incr runNo

                } ;# Score coverage thresholds

            } ;# Coverage in both sequences yes/no

        } ;# Coverage fractions

        file delete $neighboursfile

    } ;# Word sizes

    # Print results
    puts "Taxon\t[join $scores \t]"
    foreach taxon [gettaxa $bcList] {
        puts -nonewline $taxon
        for {set partition 0} {$partition < $runNo} {incr partition} {
            puts -nonewline "\t$results($taxon,$partition)"
        }
        puts ""
    }

} problem]} {
    puts stderr $problem
    exit 1
}

